"""Console script for globality_black."""
import multiprocessing as mp
import sys
from functools import partial
from pathlib import Path
from typing import Tuple

import click

from globality_black.black_handler import get_black_mode
from globality_black.constants import (
    ALL_DONE_STRING,
    NUM_FILES_TO_ENABLE_PARALLELIZATION,
    OH_NO_STRING,
)
from globality_black.reformat_text import BlackError, reformat_text


@click.command()
@click.argument("path", type=click.Path(readable=True, writable=True, exists=True))
@click.option("--check/--no-check", type=bool, default=False)
@click.option("--verbose/--no-verbose", type=bool, default=False)
# characters \b needed to avoid click reformatting
# see https://click.palletsprojects.com/en/7.x/documentation/#preventing-rewrapping
def main(path, check, verbose):
    """
    Run globality-black for a given path

    \b
    * path:
        If path is a directory, apply to all .py files in any subdirectory
        Otherwise, apply just to the given filename.

    \b
    * check:
        If --check is passed, do not modify the files and return:
            - exit code 1: if any file needs to be reformatted (or fails when applying black)
            - exit code 0: otherwise
        \b
        If --check not passed (or --no-check is passed), attempt to reformat all paths returning
            - exit code 1: if any file fails
            - exit code 0: otherwise
        \b
        Note that when not passing --check, all files not failing will be correctly reformatted
        (i.e. globality-black is independently applied per-file)

    \b
    * verbose:
        If --verbose not passed (or --no-verbose), only files with errors or to be modified are
        shown

    """

    path = Path(path)
    exit_code = 0

    if path.is_dir():
        paths = list(path.glob("**/*.py"))
    else:
        paths = [path]

    reformatted_count, failed_count = 0, 0
    process_path_with_check = partial(process_path, check_only_mode=check)

    parallelize = len(paths) > NUM_FILES_TO_ENABLE_PARALLELIZATION
    if parallelize:
        with mp.Pool(mp.cpu_count() - 1) as pool:
            map_result = pool.map(process_path_with_check, paths)
    else:
        # Do not parallelize if just a few files
        map_result = map(process_path_with_check, paths)

    for is_modified, is_failed, message in map_result:
        if verbose or is_modified or is_failed:
            click.echo(message)
        reformatted_count += is_modified
        failed_count += is_failed

    unchanged_count = len(paths) - reformatted_count - failed_count

    # add a separator line
    click.echo("-" * len(OH_NO_STRING))

    # if we are just checking and at least one file needs to be reformatted OR some file failed
    if (check and reformatted_count > 0) or failed_count > 0:
        click.echo(OH_NO_STRING)
        exit_code = 1
        if failed_count > 0:
            click.echo(f"{failed_count} files failed to parse (black error)")
    else:
        click.echo(ALL_DONE_STRING)

    if check:
        if reformatted_count > 0:
            click.echo(f"{reformatted_count} files would be reformatted")
        if unchanged_count > 0:
            click.echo(f"{unchanged_count} files would be left unchanged")
    else:
        if reformatted_count > 0:
            click.echo(f"{reformatted_count} files reformatted")
        if unchanged_count > 0:
            click.echo(f"{unchanged_count} files unchanged")

    sys.exit(exit_code)


def process_path(
    path: Path,
    check_only_mode: bool = False,
) -> Tuple[bool, bool, str]:
    """
    For each path compute `is_modified`, `is_failed`, and `message` to be used in main
    """

    is_modified = False
    input_code = path.read_text()
    black_mode = get_black_mode(path)

    try:
        output_code = reformat_text(input_code, black_mode)
    except BlackError as e:
        return False, True, f"Failed to reformat {path}. {e}"

    if input_code != output_code:
        is_modified = True

    if check_only_mode and is_modified:
        initial_str = "Would reformat"
    elif not check_only_mode and is_modified:
        initial_str = "Reformatted"
    else:
        initial_str = "Nothing to do for"

    if not check_only_mode:
        path.write_text(output_code)

    return is_modified, False, f"{initial_str} {path}"


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import re
from typing import List, Optional

from parso.python.tree import Module

from globality_black.constants import (
    MAX_CHARACTERS_TO_FIND_INDENTATION_PARENT,
    TYPES_TO_CHECK_FMT_ON_OFF,
)


class SyntaxTreeVisitor:
    def __init__(
        self,
        module: Module,
        types_to_find: Optional[List[str]] = None,
    ):
        self.module = module
        self.types_to_find = types_to_find
        self.fmt_off = False

    def __call__(self, node):

        self.set_fmt_on_off_according_to_prefix(node)
        if self.types_to_find is None or node.type in self.types_to_find and not self.fmt_off:
            yield node

        if hasattr(node, "children"):
            for child in node.children:
                self.set_fmt_on_off_according_to_prefix(child)
                if not self.fmt_off:
                    yield from self(child)

    def set_fmt_on_off_according_to_prefix(self, node):
        """
        We check for fmt_on_off only for some specific types, otherwise no-op. This saves time, so
        we don't need to find the prefix for every single node in the parse tree
        """

        if not any(name in node.type for name in TYPES_TO_CHECK_FMT_ON_OFF):
            return
        prefix = node.get_first_leaf().prefix

        if "fmt: off" in prefix:
            self.fmt_off = True

        if "fmt: on" in prefix:
            self.fmt_off = False


def apply_function_to_tree_prefixes(module, root, function):
    visitor = SyntaxTreeVisitor(module)

    for node in visitor(root):
        prefix = node.get_first_leaf().prefix
        node.get_first_leaf().prefix = function(prefix)


def find_indentation_parent_prefix(element):
    """
    Find prefix for the indentation parent by going to the parent's line and getting the indent
    for the first element in the line, i.e. the node we have to align this element with

    Examples:

          x = foo(arg1="marc",) --> indentation parent for arg1 is x (his grand-grand-parent)
          foo(arg1="marc",) --> indentation parent for arg1 is foo (his grand-parent)

    """

    parent = element.parent
    module = element.get_root_node()

    line_start_pos = (parent.start_pos[0], 0)
    leaf = module.get_leaf_for_position(line_start_pos, include_prefixes=True)

    # we move the "pointer" to position 0 of this line and check if the leaf.type is not a newline
    # otherwise we keep moving the pointer until we find something, and that gives as the
    # indentation sized we're looking for
    while leaf.type == "newline":
        leaf = module.get_leaf_for_position(line_start_pos, include_prefixes=True)
        line_start_pos = (line_start_pos[0], line_start_pos[1] + 1)

        if line_start_pos[1] > MAX_CHARACTERS_TO_FIND_INDENTATION_PARENT:
            raise ValueError(
                "Warning: no leaf found in this line after "
                f"{MAX_CHARACTERS_TO_FIND_INDENTATION_PARENT} characters. "
                f"Looking for parent for leaf:\n{leaf.get_code()}"
            )

    return leaf.prefix


def get_indent_from_prefix(prefix):
    """
    Each element in parso has a prefix. We want to get the indent from the parent, so we can
    construct the prefix for the modified elements.

    Example:

         def foo(
            arg1="marc", arg2=" at ", arg3="globality",
        ):
            return arg1 + arg2 + arg3

    arg1 has prefix "\n    ". Here we get the "    ".
    """

    if prefix:
        try:
            return re.search("( *)$", prefix).group(0)
        except Exception:
            raise ValueError(f"Could not get indent from prefix {prefix}")
    else:
        return prefix



"""
Module to explode comprehensions:
- all dict comprehensions
- all comprehensions with `if`

"""
from parso.python.tree import PythonNode

from globality_black.common import find_indentation_parent_prefix, get_indent_from_prefix
from globality_black.constants import ParsoTypes


def reformat_comprehension(comp_for: PythonNode):
    """
    comp_for represents a subset of the comprehension, e.g. in

    [(3, i): "globality" for i in range(4) if i < 2]

    comp_for is a PythonNode with type "sync_comp_for" and code
        "for i in range(4) if i < 2"

    comp is a PythonNode with type "testlist_comp" and code
        "(3, i): "globality" for i in range(4) if i < 2"

    We explode if at least one of the following requirements are met:
        - dict comprehension
        - there is an if
        - there is a ternary operator
        - there are multiple for's

    We don't explode:
        - comprehension inside of comprehension (the inner one will be kept as is), see README for
        an example
        - if multiple for's, the first one already explodes the rest
        (with `set_prefix_for_all_last_children`), so they are ignored
        Example: Given [i for i in range(4) for j in range(6)], i-for contains the j-for as
        children, so we can ignore the j-for because it will already be exploded once we reach it
    """

    assert comp_for.type == ParsoTypes.SYNC_COMP_FOR.value

    comp = comp_for.parent

    is_dict = comp.type == ParsoTypes.DICTORSETMAKER.value and comp.children[1] == ":"
    ends_with_if = comp_for.children[-1].type == ParsoTypes.COMP_IF.value
    ends_with_for = comp_for.children[-1].type == ParsoTypes.SYNC_COMP_FOR.value
    value_is_ternary_expression = comp.children[0].type == ParsoTypes.TERNARY_EXPRESSION.value
    value_is_comprehension = find_if_value_is_comprehension(comp)

    parent_is_for = comp.type == ParsoTypes.SYNC_COMP_FOR.value
    comprehension_types = (ParsoTypes.DICTORSETMAKER.value, ParsoTypes.LISTCOMP.value)
    nested_comp = comp.parent.parent.type in comprehension_types

    requirements = is_dict or ends_with_if or ends_with_for
    requirements = requirements or value_is_ternary_expression or value_is_comprehension
    can_be_ignored = parent_is_for or nested_comp

    if requirements and not can_be_ignored:
        _reformat_comprehension(comp_for)


def find_if_value_is_comprehension(comp):
    value = comp.children[0]
    for child in getattr(value, "children", []):
        if child.type == "testlist_comp":
            return True
    return False


def _reformat_comprehension(comp_for: PythonNode):
    """
    Here we do the actual reformatting
    """
    comp = comp_for.parent
    prefix = find_indentation_parent_prefix(comp)
    base_indent = get_indent_from_prefix(prefix)
    new_prefix = "\n" + base_indent + " " * 4

    # indent element of comp
    set_prefix(comp, new_prefix)

    set_prefix_for_all_last_children(comp_for, new_prefix)

    # set closing bracket
    set_prefix(comp.parent.children[-1], "\n" + base_indent)


def set_prefix(element: PythonNode, prefix: str):

    leaf = element.get_first_leaf()

    # if there's only space, replace
    if not leaf.prefix.strip():
        leaf.prefix = prefix

    # otherwise, add prefix to previous prefix (to keep comments)
    else:
        last_eol_position = leaf.prefix.rindex("\n")
        leaf.prefix = leaf.prefix[:last_eol_position] + prefix


def set_prefix_for_all_last_children(comp_for: PythonNode, prefix: str):
    """
    indent for in comprehension + all for and if underneath
    unfortunately parso treats each new comp_for and comp_if after the comp_for (if any) as a child
    of the
    previous one. Hence, we need to recursively iterate on them

    """

    set_prefix(comp_for, prefix)

    last_child = comp_for.children[-1]
    if last_child.type in {ParsoTypes.COMP_IF.value, ParsoTypes.SYNC_COMP_FOR.value}:
        set_prefix_for_all_last_children(last_child, prefix)



import black
import parso

from globality_black.blank_lines import cover_blank_lines, uncover_blank_lines
from globality_black.common import SyntaxTreeVisitor
from globality_black.comprehensions import reformat_comprehension
from globality_black.constants import BLANK_LINES_TYPES, COMPREHENSIONS_TYPES


class BlackError(Exception):
    pass


def reformat_text(file_contents, black_mode):

    module = parso.parse(file_contents)

    # PRE-PROCESSING

    finder = SyntaxTreeVisitor(module, BLANK_LINES_TYPES)

    for element in finder(module):

        # cover blank lines if needed
        cover_blank_lines(module, element)

    # BLACK

    try:
        code_after_black = black.format_str(module.get_code(), mode=black_mode)
    except Exception as e:
        raise BlackError(e)

    module = parso.parse(code_after_black)

    # POST-PROCESSING

    # comprehensions
    finder = SyntaxTreeVisitor(module, COMPREHENSIONS_TYPES)
    for element in finder(module):
        if element.type == "sync_comp_for":
            reformat_comprehension(element)

    # uncover blank lines protected during pre-processing
    finder = SyntaxTreeVisitor(module, BLANK_LINES_TYPES)
    for element in finder(module):
        if element.type in BLANK_LINES_TYPES:
            uncover_blank_lines(module, element)

    return module.get_code()



"""
Module to cover/uncover specific blank lines in pre/post processor
 so that black does not remove them

The blank lines we want to cover are those added explicitly by the developer to improve
readability, e.g. blank lines below:

graph.use(
    "logging",

    "space",

    # Sagemaker basics
    "sagemaker",

    # This line causes ...
    "load_active_bundle_and_dependencies",
)
"""
import re

from globality_black.common import apply_function_to_tree_prefixes
from globality_black.constants import BLANK_LINE_TOKEN


def cover_blank_lines(module, root):
    apply_function_to_tree_prefixes(module, root, add_token_if_line_to_keep)


def add_token_if_line_to_keep(prefix):
    # Add a token when two new lines

    return re.sub(
        r"\n{2}( +)",
        fr"\n# {BLANK_LINE_TOKEN}\n\1",
        prefix,
    )


def uncover_blank_lines(module, root):
    apply_function_to_tree_prefixes(
        module,
        root,
        remove_token_from_covered_line,
    )


def remove_token_from_covered_line(prefix):
    # Note that black will add spaces before the token

    return re.sub(
        fr"\n\s+# {BLANK_LINE_TOKEN}\n",
        r"\n\n",
        prefix,
    )



from enum import Enum, unique


BLANK_LINE_TOKEN = "TOKEN"
BLANK_LINES_TYPES = ["arglist", "expr_stmt", "atom_expr", "atom"]
COMPREHENSIONS_TYPES = ["comp_if", "sync_comp_for"]
TYPES_TO_CHECK_FMT_ON_OFF = ("stmt", "funcdef", "classdef")
MAX_CHARACTERS_TO_FIND_INDENTATION_PARENT = 200
DEFAULT_BLACK_LINE_LENGTH = 100
NUM_FILES_TO_ENABLE_PARALLELIZATION = 5
ALL_DONE_STRING = "All done! ✨ 🍰 ✨"
OH_NO_STRING = "Oh no! 💥 💔 💥"


@unique
class ParsoTypes(Enum):
    DICTORSETMAKER = "dictorsetmaker"
    COMP_IF = "comp_if"
    SYNC_COMP_FOR = "sync_comp_for"
    TERNARY_EXPRESSION = "test"
    LISTCOMP = "testlist_comp"



from functools import lru_cache
from pathlib import Path

import black

from globality_black.constants import DEFAULT_BLACK_LINE_LENGTH


@lru_cache(maxsize=1)
def get_black_mode(src: Path) -> black.Mode:
    """Read the black configuration from pyproject.toml"""

    value = black.find_pyproject_toml((str(src),))

    if not value:
        return black.Mode(line_length=DEFAULT_BLACK_LINE_LENGTH)

    config = black.parse_pyproject_toml(value)

    return black.Mode(
        **{
            key: value
            for key, value in config.items()
            if key in ["line_length", "skip_string_normalization"]
        }
    )



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup


with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

project = "globality-black"
version = "2021.10.444"

setup(
    name=project,
    author="Globality AI",
    version=version,
    author_email="ai@globality.com",
    python_requires=">=3.7",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="A wrapper for black adding new features",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    url="https://github.com/globality-corp/globality-black",
    keywords="globality_black",
    entry_points={
        "console_scripts": [
            "globality-black = globality_black.cli:main",
        ],
    },
    install_requires=[
        "Click>=7.0",
        "parso",
        "pytest>=3",
        "black==20.8b1",
    ],
    extras_require={
        "lint": [
            "flake8-isort>=3.0.1",
            "flake8-print>=3.1.0",
            "flake8-logging-format",
            "globality-black",
        ],
        "test": [
            "pytest",
            "pytest-cov",
        ],
    },
    include_package_data=True,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    setup_requires=[
        "pytest-runner",
    ],
    test_suite="tests",
    zip_safe=False,
)
